from itertools import chain, islice
import networkx as nx
import pygna.reading_class as rc

def parse_network(network: str, max_components=None,
                  min_component_size: int = 2) -> nx.Graph:
    """Load network and geneset files, and return the network or subnetwork
    to operate on

    Parameters
    ----------
    network : str
        TSV file defining the input network
    max_components
        Load only n largest components
    min_component_size
        Load only components with at least n genes (default: 2)

    Returns
    -------
    nx.Graph
        The [filtered] network
    """

    network = rc.ReadTsv(network).get_network()
    if max_components or (min_component_size > 2):
        filtered_network = nx.Graph()
        for component in filter(lambda c: len(c)>=min_component_size,
                                islice(sorted(nx.connected_components(network),
                                    key=len, reverse=True), max_components)):
            filtered_network = nx.compose(filtered_network,
                                          network.subgraph(component).copy())
        network = filtered_network
    return network
    

def parse_genesets(genesets: str, setnames=None, min_set_size: int = 20) -> dict:
    """Load gene set files and parse them into a dict. If setnames are given,
    load the named sets, otherwise load all sets in the file

    Parameters
    ----------
    genesets
        GMT file containing gene sets (default: None)
    setnames
        List of gene set names (default: None)

    Returns
    -------
    dict
        Dictiinary of gene sets
    """

    if not setnames:
        setnames = tuple(gs for gs in rc.ReadGmt(genesets).get_data().keys()
                            if gs != 'Gene Set')
    genesets = dict((setname, genes) for setname, genes in chain.from_iterable(
        rc.ReadGmt(genesets).get_geneset(setname).items()
        for setname in setnames) if len(genes) > min_set_size)
    return genesets


def parse_network_and_genesets(network, genesets=None,
                               setnames=None, max_components=None,
                               min_component_size: int = 2,
                               min_set_size: int = 20) -> (nx.Graph, dict):
    """Load network and geneset files, and return the network or subnetwork
    to operate on

    Parameters
    ----------
    network : str
        TSV file defining the input network
    genesets
        GMT file containing gene sets (default: None)
    setnames
        List of gene set names (default: None)
    max_components
        Load only n largest components
    min_component_size : int
        Load only components with at least n genes (default: 2)
    min_set_size : int
        Only consider gene sets with more than n genes

    Returns
    -------
    nx.Graph
        The network of interest
    dict
        Dictionary of gene sets
    """

    network = parse_network(network, max_components=max_components,
                            min_component_size=min_component_size)
    genesets = parse_genesets(genesets, setnames=setnames,
                              min_set_size=min_set_size) if genesets else {}
    return network, genesets


import pandas as pd
import networkx as nx
import logging
from math import log10, floor, ceil
from statistics import mean, variance, stdev
from itertools import combinations, product
from multiprocessing import Pool
from functools import partial
import pygna.statistical_comparison as sc
from pygna2.parse_network_genesets import parse_network, parse_genesets
from pygna2.test_topology import rw_dict
from pygna2.test_plot import (plot_test_results, plot_test_clustermap,
                              plot_test_heatmap)

def run_comparison(sc_comparison, genes_a, genes_b, permutations: int = 200, cores: int = 1):
    (observed, pvalue, null_d, a_mapped, b_mapped) = sc_comparison.comparison_empirical_pvalue(
        genes_a, genes_b, max_iter=permutations, alternative="greater", keep=True
    )
    return observed, pvalue, null_d


def run_comparison_on_genesets(sc_comparison, setname_a, genes_a, setname_b, genes_b,
                                permutations: int = 200,
                                processes: int = 1, verbose=False):
    if verbose:
        logging.info(f'Running test on {setname_a} vs {setname_b}')
    (observed, pvalue, null_d) = run_comparison(sc_comparison, genes_a, genes_b,
        permutations=permutations, cores=processes)
    if verbose:
        logging.info(f'Observed value: {observed}, P-value: {pvalue}')
    return setname_a, setname_b, observed, pvalue, null_d


def test_association(network: str, genesets_a: str, output: str,
                     rwr_matrix: str, genesets_b=None,
                     setnames_a=None, setnames_b=None, max_components=None,
                     min_component_size: int = 2,
                     min_set_size: int = 20, results_figure=None,
                     results_figure_type: str = 'box',
                     results_figure_significance: bool = False,
                     width: float = 7, height: float = 7,
                     null_distributions=None,
                     filtered_network_file=None,
                     degree_correction_bins=1,
                     rwr_matrix_in_memory=False,
                     permutations: int = 100, processes: int = 1):
    network  = parse_network(network, max_components=max_components,
        min_component_size=min_component_size)
    genesets_a = parse_genesets(genesets_a, setnames=setnames_a,
                                min_set_size=min_set_size)
    if genesets_b:
        genesets_b = parse_genesets(genesets_b, setnames=setnames_b,
                                    min_set_size=min_set_size)
    sc_comparison = sc.StatisticalComparison(sc.comparison_random_walk,
        network, n_proc=processes, degree_bins=degree_correction_bins,
        diz=rw_dict(rwr_matrix, in_memory=rwr_matrix_in_memory))
    geneset_pairs = tuple((setname_a, genes_a, setname_b, genes_b)
                          for (setname_a, genes_a), (setname_b, genes_b)
                          in (combinations(genesets_a.items(), 2)
                              if not genesets_b else product(genesets_a.items(),
                                                             genesets_b.items()))
                          if setname_a != setname_b)
    results = tuple(run_comparison_on_genesets(sc_comparison, setname_a,
        genes_a, setname_b, genes_b, permutations=permutations,
        processes=processes, verbose=True)
        for setname_a, genes_a, setname_b, genes_b in geneset_pairs)
    pd.DataFrame(
        (('association_rwr', setname_a, setname_b, permutations, observed,
          pvalue, mean(null_d), variance(null_d))
         if len(set(null_d))>1 else ('association_rwr', setname_a, setname_b,
                                     permutations, observed, pvalue,
                                     mean(null_d), float('inf'))
         for setname_a, setname_b, observed, pvalue, null_d in results),
         columns=('Analysis', 'Gene set A', 'Gene set B', 'Permutations',
                  'Observed', 'P-value', 'Null mean', 'Null variance')
    ).to_csv(output, index=False, sep='\t' if output.endswith('.tsv') else ',')
    null_dists = pd.DataFrame(
        ((setname_a, setname_b, val) for setname_a, setname_b, _, _, null_d
         in results for val in null_d),
        columns=('Gene set A', 'Gene set B', 'Null distribution')
    )
    observed = pd.DataFrame(
        ((setname_a, setname_b, observed, pvalue,
          (observed - mean(null_d)) / stdev(null_d),
          '*'*min(3, floor(-1*log10(2*pvalue))))
         for setname_a, setname_b, observed, pvalue, null_d in results),
        columns=('Gene set A', 'Gene set B', 'Observed', 'P-value', 'Z-score',
                 'Significance')
    )
    if results_figure:
        if not genesets_b:
            plot_test_clustermap(observed, results_figure,
                              figure_type=results_figure_type.replace('box', 'obs'),
                              significance=results_figure_significance,
                              width=width, height=height)
        elif len(genesets_b) == 1 or len(genesets_a) == 1:
            if len(genesets_b) == 1:
                null_dists['Gene set'] = null_dists['Gene set A']
                observed['Gene set'] = observed['Gene set A']
            elif len(genesets_a) == 1:
                null_dists['Gene set'] = null_dists['Gene set B']
                observed['Gene set'] = observed['Gene set B']
            plot_test_results(null_dists,
                              observed.loc[:,['Gene set', 'Observed', 'Significance']],
                              results_figure,
                              figure_type=results_figure_type,
                              significance=results_figure_significance,
                              width=width, height=height,
                              strip_alpha=min(1, 32/permutations),
                              x_label='Association RWR statistic')
        else:
            plot_test_heatmap(observed, results_figure,
                              figure_type=results_figure_type.replace('box', 'obs'),
                              significance=results_figure_significance,
                              width=width, height=height)
    if null_distributions:
        null_dists.to_csv(null_distributions, index=False,
                          sep='\t' if null_distributions.endswith('.tsv')
                          else ',')
    if filtered_network_file:
        nx.to_pandas_edgelist(network).to_csv(filtered_network_file,
            sep='\t', header=False, index=False)


#===============================================================================
# matrix.py
#===============================================================================

"""Build the RWR diffusion matrix"""

# Imports ======================================================================

import networkx as nx
import numpy as np
import scipy
import logging
import tables
from pygna2.parse_network_genesets import parse_network

# Functions ====================================================================

def build_rwr_diffusion_matrix(network, output, beta: float = 0.85,
                               max_components=None,
                               min_component_size: int = 2,
                               filtered_network_file=None):
    """Build the RWR diffusion matrix
    
    Parameters
    ----------
    network : str
        TSV file defining the input network
    output : str
        Text file containing summary
    beta : float
        Beta (restart probability) parameter of diffusion
    max_components
        Load only n largest components
    min_component_size
        Load only components with at least n genes (default: 2)
    filtered_network
        Output file for filtered network (default: None)
    """

    network = parse_network(network, max_components=max_components,
                            min_component_size=min_component_size)
    nodes = list(network.nodes())
    logging.info("Beginning to calculate RWR matrix")
    a = nx.adjacency_matrix(network)
    k = 1 / np.array(list(dict(network.degree()).values()))
    d = scipy.sparse.dia_matrix((k, [0]), shape=a.shape)
    a = a.dot(d)
    n = np.shape(a)[1]
    with tables.open_file(output, mode="w") as hdf5_file:
        # create a hdf5 file with two objects:
        # - one is the nodes array,
        hdf5_file.create_array(hdf5_file.root, "nodes", nodes)
        # -  the other is the RWR matrix
        hdf5_file.create_array(hdf5_file.root, "matrix",
                               beta*np.linalg.inv(np.eye(n)-(1.0-beta)*a))
        logging.info("Saving network")
        hdf5_file.close()
    if filtered_network_file:
        nx.to_pandas_edgelist(network).to_csv(filtered_network_file,
            sep='\t', header=False, index=False)



#===============================================================================
# network.py
#===============================================================================

"""Network properties"""

# Imports ======================================================================

import logging
import numpy as np
from itertools import chain, combinations, islice
import networkx as nx
from pygna.output import write_graph_summary
from pygna.diagnostic import plot_connected_components, plot_degree
from pygna2.parse_network_genesets import parse_network_and_genesets

# Functions ====================================================================

def filter_by_genes(network: nx.Graph, genes: set, strict: bool = False,
                       minimal: bool = False):
    network_nodes = network.nodes()
    if strict:
        return nx.subgraph(network, genes)
    else:
        filtered_network = nx.Graph()
    if minimal:
        for source, target in combinations((g for g in genes
                                            if g in network_nodes), 2):
            path = nx.shortest_path(network, source=source, target=target)
            if len(path) < np.inf:
                filtered_network.add_path(path)
    else:
        for component in (network.subgraph(c).copy()
                        for c in nx.connected_components(network)):
            if set(component.nodes()).intersection(genes):
                filtered_network = nx.compose(filtered_network, component)
    return filtered_network


def generate_filtered_components(network, max_components=None,
                                 min_component_size: int = 2):
    yield from filter(lambda c: len(c)>=min_component_size,
        islice(sorted((nx.connected_components(network)),
                        key=len, reverse=True), max_components))


def filter_by_components(network, max_components=None,
                         min_component_size: int = 2):
    filtered_network = nx.Graph()
    for component in generate_filtered_components(network,
        max_components=max_components, min_component_size=min_component_size):
        filtered_network = nx.compose(filtered_network,
                                      network.subgraph(component).copy())
    return filtered_network


def summary(network: str, output: str, genesets=None, setnames=None,
            net_name: str = 'network', max_components=None,
            min_component_size: int = 2, min_set_size: int = 20,
            minimal: bool = False, strict: bool = False, plot_degree_file=None,
            plot_component_file=None, filtered_network_file=None):

    """Generate a text file summarizing either the entire input network or
    a subnetwork defined by one or more gene sets.

    Parameters
    ----------
    network : str
        TSV file defining the input network
    output : str
        Text file containing summary
    genesets
        GMT file containing gene sets (default: None)
    setnames
        List of gene set names (default: None)
    min_set_size : int
        Only consider gene sets with more than n genes (default: 20)
    minimal : bool
        Filter out genes that are not part of a path between set members (default: False)
    strict : bool
        Filter out genes that are not part of a gene set (default: False)
    net_name : str
        Name of network to record in output file (default: network)
    max_components
        Load only n largest components
    min_component_size
        Load only components with at least n genes (default: 2)
    max_output_components
        Store only n largest components
    min_output_component_size
        Store only components with at least n genes (default: 2)
    plot_degree_file
        Output file for degree plot (default: None)
    plot_component_file
        Output file for component plot (default: None)
    filtered_network_file
        Output file for filtered network (default: None)
    """
    
    logging.info("Evaluating network summary, please wait")
    network, genesets = parse_network_and_genesets(network, genesets=genesets,
        setnames=setnames, max_components=max_components,
        min_component_size=min_component_size, min_set_size=min_set_size)
    if genesets:
        genes = set(chain.from_iterable(genesets.values()))
        network = filter_by_genes(network, genes, strict=strict,
                                  minimal=minimal)
    write_graph_summary(network, output, net_name=net_name)
    if plot_degree_file:
        plot_degree(nx.degree(network), plot_degree_file)
    if plot_component_file:
        plot_connected_components(nx.connected_components(network),
                                  plot_component_file)
    if filtered_network_file:
        nx.to_pandas_edgelist(network).to_csv(filtered_network_file,
            sep='\t', header=False, index=False)
    logging.info("Network summary completed")


def filter_network(network: str, output: str, genesets=None, setnames=None,
           max_components=None, min_component_size: int = 2,
           min_set_size: int = 20, minimal: bool = False, strict: bool = False,
           max_output_components=None, min_output_component_size: int = 2):
    
    """Filter small components out of the network. By default, when gene sets
    are provided, the network is filtered to components that overlap with

    Parameters
    ----------
    network : str
        TSV file defining the input network
    output : str
        TSV file containing filtered network
    genesets
        GMT file containing gene sets (default: None)
    setnames
        List of gene set names (default: None)
    min_set_size : int
        Only consider gene sets with more than n genes (default: 20)
    minimal : bool
        Filter out genes that are not part of a path between set members (default: False)
    strict : bool
        Filter out genes that are not part of a gene set (default: False)
    max_components
        Load only n largest components
    min_component_size
        Load only components with at least n genes (default: 2)
    max_output_components
        Store only n largest components
    min_output_component_size
        Store only components with at least n genes (default: 2)
    """

    network, genesets = parse_network_and_genesets(network, genesets=genesets,
        setnames=setnames, max_components=max_components,
        min_component_size=min_component_size, min_set_size=min_set_size)
    if genesets:
        genes = set(chain.from_iterable(genesets.values()))
        network = filter_by_genes(network, genes, strict=strict,
                                  minimal=minimal)
    if max_output_components or (min_output_component_size > 2):
        network = filter_by_components(network,
            max_components=max_output_components,
            min_component_size=min_output_component_size)
    nx.to_pandas_edgelist(network).to_csv(output,
        sep='\t', header=False, index=False)

def cytoscape(network: str, output_graphml: str, components_out=None,
              genesets=None, setnames=None,
              max_components=None, min_component_size: int = 2,
              min_set_size: int = 20, strict: bool = False,
              minimal: bool = False, full_network: bool = False,
              max_output_components=None, min_output_component_size: int = 2,
              filtered_network_file=None):

    """Generate a text file summarizing either the entire input network or
    a subnetwork defined by one or more gene sets.

    Parameters
    ----------
    network : str
        TSV file defining the input network
    output_graphml : str
        GraphML file containing annotated network or subnetwork
    components_out
        Write GMT file containing components of the output graph
    genesets
        GMT file containing gene sets (default: None)
    setnames
        List of gene set names (default: None)
    max_components
        Load only n largest components
    min_component_size
        Load only components with at least n genes (default: 2)
    min_set_size : int
        Only consider gene sets with more than n genes (default: 20)
    max_output_components
        Store only n largest components
    min_output_component_size
        Store only components with at least n genes (default: 2)
    minimal : bool
        Filter out genes that are not part of a path between set members (default: False)
    strict : bool
        Filter out genes that are not part of a gene set (default: False)
    full_network : bool
        Record the full network (default: False)
    filtered_network_file
        Output file for filtered network (default: None)
    """

    network, genesets = parse_network_and_genesets(network, genesets=genesets,
        setnames=setnames, min_set_size=min_set_size,
        max_components=max_components, min_component_size=min_component_size)
    if not full_network:
        if not genesets:
            raise RuntimeError('Filtering witout genesets will produce an empty output, consider using --full')
        genes = set(chain.from_iterable(genesets.values()))
        network = filter_by_genes(network, genes, strict=strict,
                                  minimal=minimal)
    network_nodes = network.nodes()
    for setname, genes in genesets.items():
        dict_nodes = {n: n in genes for n in network_nodes}
        nx.set_node_attributes(network, dict_nodes, setname)
    nx.write_graphml(network, output_graphml)
    if components_out:
        with open(components_out, 'w') as f:
            for index, component in enumerate(generate_filtered_components(
                network, max_components=max_output_components,
                min_component_size=min_output_component_size)):
                f.write('\t'.join((f'component_{index}', str(len(component)))
                                  + tuple(component)) + '\n')
    if filtered_network_file:
        if max_output_components or (min_output_component_size > 2):
            network = filter_by_components(network,
                max_components=max_output_components,
                min_component_size=min_output_component_size)
        nx.to_pandas_edgelist(network).to_csv(filtered_network_file,
            sep='\t', header=False, index=False)



import pandas as pd
import networkx as nx
import logging
from math import log10, floor
from statistics import mean, variance
from multiprocessing import Pool
from functools import partial
import pygna.statistical_test as st
from pygna.command import read_distance_matrix

from pygna2.parse_network_genesets import parse_network_and_genesets
from pygna2.test_plot import plot_test_results

TABLE_HEADER = ("analysis,setname,n_mapped,n_geneset,number_of_permutations,observed,empirical_pvalue,mean(null),"
                      "var(null),network,geneset\n")

def run_test(st_test, genes, permutations: int = 200, cores: int = 1):
    (observed, pvalue, null_d, n_mapped, n_geneset) = st_test.empirical_pvalue(
        genes, max_iter=permutations, alternative="greater", cores=cores
    )
    return observed, pvalue, null_d


def rw_dict(rwr_matrix, in_memory=False):
    if rwr_matrix:
        return dict(zip(('nodes', 'matrix'),
            read_distance_matrix(rwr_matrix, in_memory=in_memory)))


def run_test_on_geneset(st_test, setname, genes, permutations: int = 200,
                        processes: int = 1, verbose=False):
    if verbose:
        logging.info(f'Running test on {setname}')
    (observed, pvalue, null_d) = run_test(st_test, genes,
        permutations=permutations, cores=processes)
    if verbose:
        logging.info(f'Observed value: {observed}, P-value: {pvalue}')
    return setname, observed, pvalue, null_d


def test_topology(network: str, genesets: str, output: str, rwr_matrix=None,
                  stat=st.geneset_internal_degree_statistic,
                  stat_name='topology_internal_degree',
                  setnames=None, max_components=None,
                  min_component_size: int = 2,
                  min_set_size: int = 20, results_figure=None,
                  results_figure_type: str = 'box',
                  results_figure_significance: bool = False,
                  results_figure_x_label: str = 'Topology Internal Degree (%)',
                  width: float = 7, height: float = 7,
                  null_distributions=None,
                  filtered_network_file=None,
                  degree_correction_bins=1,
                  rwr_matrix_in_memory=False,
                  permutations: int = 200, processes: int = 1):
    network, genesets = parse_network_and_genesets(network, genesets=genesets,
        setnames=setnames, max_components=max_components,
        min_component_size=min_component_size, min_set_size=min_set_size)
    st_test = st.StatisticalTest(stat, network,
        degree_bins=degree_correction_bins,
        diz=rw_dict(rwr_matrix, in_memory=rwr_matrix_in_memory)
                    if rwr_matrix else {})
    results = tuple(run_test_on_geneset(st_test, setname, genes,
                                        permutations=permutations,
                                        processes=processes, verbose=True)
                    for setname, genes in genesets.items())
    pd.DataFrame(
        ((stat_name, setname, permutations, observed, pvalue, mean(null_d),
          variance(null_d))
         if len(set(null_d))>1 else (stat_name, setname, permutations,
                                     observed, pvalue, mean(null_d),
                                     float('inf'))
         for setname, observed, pvalue, null_d in results
        ),
         columns=('Analysis', 'Gene set', 'Permutations', 'Observed',
                  'P-value', 'Null mean', 'Null variance')
    ).to_csv(output, index=False, sep='\t' if output.endswith('.tsv') else ',')
    null_dists = pd.DataFrame(
        ((setname, val) for setname, _, _, null_d in results for val in null_d),
        columns=('Gene set', 'Null distribution')
    )
    observed = pd.DataFrame(
        ((setname, observed, '*'*min(3, floor(-1*log10(2*pvalue))))
         for setname, observed, pvalue, _ in results),
        columns=('Gene set', 'Observed', 'Significance')
    )
    if results_figure:
        plot_test_results(null_dists, observed, results_figure,
                          figure_type=results_figure_type,
                          significance=results_figure_significance,
                          width=width, height=height,
                          strip_alpha=min(1, 32/permutations),
                          x_label=results_figure_x_label)
    if null_distributions:
        null_dists.to_csv(null_distributions, index=False,
                          sep='\t' if null_distributions.endswith('.tsv')
                          else ',')
    if filtered_network_file:
        nx.to_pandas_edgelist(network).to_csv(filtered_network_file,
            sep='\t', header=False, index=False)



"""PyGNA2 is based on `PyGNA <https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-020-03801-1>`_. It uses the same statistical tests, but improves the way results are reported.

PyGNA is a tool that can perform several kinds of statistical tests on gene networks and gene sets.

- Publication: https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-020-03801-1
- Documentation: https://pygna.readthedocs.io/en/latest/index.html
- Github: https://github.com/stracquadaniolab/pygna
- Workflow github: https://github.com/stracquadaniolab/workflow-pygna
"""


import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from math import log10
from scipy.cluster.hierarchy import linkage
from itertools import combinations

ANALYSIS_LABEL = {'topology_total_degree': 'Topology Total Degree',
                  'topology_internal_degree': 'Topology Internal Degree (%)',
                  'topology_module': 'Largest Module Size',
                  'topology_rwr': 'Topology RWR statistic',
                  'association_rwr': 'Association RWR statistic'}

sns.set_style('whitegrid')

def parse_results_file(results_file: str):
    sep = '\t' if results_file.endswith('.tsv') else ','
    observed = pd.read_csv(results_file, sep=sep)
    observed['Significance'] = ['*'*int(n_stars)
                                for n_stars in np.floor(
                                    -1*np.log10(2*observed['P-value'])
                                ).combine(3, min)]
    if observed.loc[0, 'Analysis'].startswith('topology'):
        results_type = 'one-v-many'
    elif len(observed['Gene set A'].unique()) == 1 or len(observed['Gene set B'].unique()) == 1:
        results_type = 'one-v-many'
    elif sorted(combinations(set(observed['Gene set A']).union(set(observed['Gene set B'])), 2)) == sorted(zip(observed['Gene set A'], observed['Gene set B'])):
        results_type = 'all-v-all'
    else:
        results_type = 'many-v-many'
    return observed, results_type

def plot_test_results(null_dists, observed, output: str,
                      figure_type: str = 'box', significance=False,
                      width: float = 7.0, height: float = 7.0,
                      strip_alpha: float = 1,
                      x_label='Topology Internal Degree (%)'):
    n_genesets = len(observed.index)
    if x_label.endswith('(%)'):
        null_dists['Null distribution'] *= 100
        observed['Observed'] *= 100
    if figure_type == 'box':
        ax = sns.boxplot(
            data=null_dists,
            x='Null distribution',
            y='Gene set',
            whis=[0, 100],
            width=.3,
            palette=sns.husl_palette(n_genesets, l=0.9),
            zorder=1
        )
    elif figure_type == 'violin':
        ax = sns.violinplot(
            data=null_dists,
            x='Null distribution',
            y='Gene set',
            inner=None,
            palette=sns.husl_palette(n_genesets, l=0.9),
            zorder=1
        )
    elif figure_type == 'strip':
        ax = sns.stripplot(
            data=null_dists,
            x='Null distribution',
            y='Gene set',
            alpha=strip_alpha,
            palette=sns.husl_palette(n_genesets, s=0.4),
            zorder=1
        )
    sns.pointplot(ax=ax, data=observed, x='Observed', y='Gene set',
                  join=False, markers='d', errorbar=None, color='white',
                  scale=1.1)
    sns.pointplot(ax=ax, data=observed, x='Observed', y='Gene set',
                  join=False, markers='d', errorbar=None, palette='husl',
                  scale=1)
    if significance:
        for tick, (_, (_, obs, sig)) in zip(ax.get_yticks(),
                                            observed.iterrows()):
            if sig:
                ax.text(obs, tick, sig, horizontalalignment='center',
                        verticalalignment='bottom', color='black')
    sns.despine(left=True, bottom=True)
    plt.xticks(rotation=30, ha='right')
    ax.set(ylabel="", xlabel=x_label)
    fig = ax.get_figure()
    fig.tight_layout()
    fig.set_figwidth(width)
    fig.set_figheight(height)
    fig.savefig(output)


def compute_linkage_matrix(stat_matrix, method='complete', optimal_ordering=True):
    """Compute a linkage matrix defining a hierarchical clustering
    
    Parameters
    ---------
    stat_matrix
        pandas data frame containing the stat matrix
    method : str
        linkage algorithm for hierarchical clustering. See
        `scipy.cluster.hierarchy.linkage` for details.
    optimal_ordering : bool
        If True, the linkage matrix will be reordered so that the distance
        between successive leaves is minimal. See
        `scipy.cluster.hierarchy.linkage`
    """

    stat_matrix_np = stat_matrix.to_numpy()
    similarity = stat_matrix_np[np.triu_indices(len(stat_matrix.index), k=1)]
    distance = 1 - similarity / similarity.max()
    return linkage(distance, method=method, optimal_ordering=optimal_ordering)


def plot_test_clustermap(observed, output: str, figure_type: str = 'obs',
                         significance=False, cmap: str = 'rocket_r',
                         width: float = 7.0, height: float = 7.0,
                         method='complete', optimal_ordering: bool = True,
                         square: bool = True, heatmap_tick_pos='right',
                         cbar_tick_pos='left', dendrogram_ratio: float = 0.2,
                         dendrogram_spacer: float = 0.1):

    """Plot a clustered heatmap of genome similarity values

    Parameters
    ----------
    observed: DataFrame
        Pandas data frame representing the observed values
    output : str
        path to destination file for plot
    cmap : str
        color map for the heatmap
    width : float
        width of the plot in inches
    height : float
        height of the plot in inches
    method : str
        linkage algorithm for hierarchical clustering. See
        `scipy.cluster.hierarchy.linkage` for details.
    optimal_ordering : bool
        If True, the linkage matrix will be reordered so that the distance
        between successive leaves is minimal. See
        `scipy.cluster.hierarchy.linkage`
    square : bool
        If True, the heatmap will be drawn square. This may force the figure
        dimensions to be altered slightly
    heatmap_tick_pos: str
        Position of heatmap ticks. Must be "left" or "right" [right]
    cbar_tick_pos : str
        Position of color bar ticks. Must be "left" or "right" [left]
    dendrogram_ratio : float
        Fraction of plot width used for dendrogram [0.2]
    dendrogram_spacer : float
        Fraction of plot width used as spacer between dendrogram and heatmap [0.1]
    """

    genesets = sorted(set(observed['Gene set A']).union(set(observed['Gene set B'])))
    stat_matrix = pd.DataFrame(dtype=float, index=genesets, columns=genesets)
    sig_matrix = pd.DataFrame('', dtype=str, index=genesets, columns=genesets)
    for _, (setname_a, setname_b, obs, pvalue, zscore, sig) in observed.iterrows():
        if figure_type == 'obs':
            stat_matrix.loc[setname_a, setname_b] = obs
            stat_matrix.loc[setname_b, setname_a] = obs
        elif figure_type == 'zscore':
            stat_matrix.loc[setname_a, setname_b] = zscore
            stat_matrix.loc[setname_b, setname_a] = zscore
        elif figure_type == 'pvalue':
            stat_matrix.loc[setname_a, setname_b] = -1 * log10(pvalue)
            stat_matrix.loc[setname_b, setname_a] = -1 * log10(pvalue)
        sig_matrix.loc[setname_a, setname_b] = -1 * log10(pvalue)
        sig_matrix.loc[setname_b, setname_a] = -1 * log10(pvalue)
        # sig_matrix.loc[setname_a, setname_b] = sig
        # sig_matrix.loc[setname_b, setname_a] = sig

    link = compute_linkage_matrix(stat_matrix, method=method,
                                     optimal_ordering=optimal_ordering)
    ax = sns.clustermap(stat_matrix, cmap=cmap, figsize=(width, height),
                        row_linkage=link, col_linkage=link,
                        xticklabels=False,
                        dendrogram_ratio=(dendrogram_ratio,0),
                        annot=sig_matrix if significance else False)
    if square:
        ax.ax_heatmap.set_aspect("equal")
    ax.ax_col_dendrogram.set_visible(False)
    ax_hm_pos = ax.ax_heatmap.get_position()
    ax_row_pos = ax.ax_row_dendrogram.get_position()
    if heatmap_tick_pos == 'left':
            ax.ax_heatmap.set_position([.9-ax_hm_pos.width, ax_hm_pos.y0,
                                        ax_hm_pos.width, ax_hm_pos.height])
    elif heatmap_tick_pos == 'right':
            ax.ax_heatmap.set_position([ax_hm_pos.x0-.1, ax_hm_pos.y0,
                                        ax_hm_pos.width, ax_hm_pos.height])
    ax.ax_heatmap.yaxis.set_ticks_position(heatmap_tick_pos)
    ax.ax_heatmap.set_yticklabels(ax.ax_heatmap.get_yticklabels(), rotation=0)
    ax.ax_row_dendrogram.set_position([ax_row_pos.x0, ax_hm_pos.y0, 
                                       ax_row_pos.width-dendrogram_spacer,
                                       ax_hm_pos.height])
    ax.ax_cbar.set_position([.97, ax_hm_pos.y0, .03, ax_hm_pos.height])
    ax.ax_cbar.yaxis.set_ticks_position(cbar_tick_pos)
    ax.savefig(output)


def plot_test_heatmap(observed, output: str, figure_type: str = 'obs',
                         significance=False, cmap: str = 'rocket_r',
                         width: float = 7.0, height: float = 7.0,
                         square: bool = True):
    genesets_a = observed['Gene set A'].unique()
    genesets_b = observed['Gene set B'].unique()
    stat_matrix = pd.DataFrame(dtype=float, index=genesets_a, columns=genesets_b)
    sig_matrix = pd.DataFrame('', dtype=str, index=genesets_a, columns=genesets_b)
    for _, (setname_a, setname_b, obs, pvalue, zscore, sig) in observed.iterrows():
        if figure_type == 'obs':
            stat_matrix.loc[setname_a, setname_b] = obs
        elif figure_type == 'zscore':
            stat_matrix.loc[setname_a, setname_b] = zscore
        elif figure_type == 'pvalue':
            stat_matrix.loc[setname_a, setname_b] = -1 * log10(pvalue)
        sig_matrix.loc[setname_a, setname_b] = -1 * log10(pvalue)
        # sig_matrix.loc[setname_a, setname_b] = sig
    ax = sns.heatmap(stat_matrix, cmap=cmap, square=square,
                     annot=sig_matrix if significance else False)
    fig = ax.get_figure()
    fig.tight_layout()
    fig.set_figwidth(width)
    fig.set_figheight(height)
    fig.savefig(output)



#===============================================================================
# pygna2.py
#===============================================================================

"""Gene network analysis"""

# Imports ======================================================================

import numpy as np
import pandas as pd
import logging
import sys
from argparse import ArgumentParser
from math import floor, log10
import pygna.statistical_test as st
from pygna2.version import __version__
from pygna2.network import summary, filter_network, cytoscape
from pygna2.matrix import build_rwr_diffusion_matrix
from pygna2.test_topology import test_topology
from pygna2.test_association import test_association
from pygna2.test_plot import (ANALYSIS_LABEL, plot_test_results,
                              plot_test_clustermap, plot_test_heatmap,
                              parse_results_file)

# Logging ======================================================================

logging.basicConfig(level=logging.INFO)

# Functions ====================================================================

def _summary(args):
    summary(args.network, args.output, genesets=args.genesets,
            setnames=args.setname, net_name=args.network_name,
            max_components=args.max_components,
            min_component_size=args.min_component_size,
            min_set_size=args.min_set_size, minimal=args.minimal,
            strict=args.strict, plot_degree_file=args.plot_degree,
            plot_component_file=args.plot_component,
            filtered_network_file=args.filtered_network)


def _filter_network(args):
    filter_network(args.network, args.output, genesets=args.genesets,
            setnames=args.setname, max_components=args.max_components,
            min_component_size=args.min_component_size,
            min_set_size=args.min_set_size, minimal=args.minimal,
            strict=args.strict,
            max_output_components=args.max_output_components,
            min_output_component_size=args.min_output_component_size)


def _cytoscape(args):
    cytoscape(args.network, args.output_graphml,
              components_out=args.components_out,
              genesets=args.genesets, setnames=args.setname,
              max_components=args.max_components,
              min_component_size=args.min_component_size,
              min_set_size=args.min_set_size, minimal=args.minimal,
              strict=args.strict, full_network=args.full,
              max_output_components=args.max_output_components,
              min_output_component_size=args.min_output_component_size,
              filtered_network_file=args.filtered_network)


def _build_rwr_diffusion_matrix(args):
    build_rwr_diffusion_matrix(args.network, args.output, beta=args.beta,
                               max_components=args.max_components,
                               min_component_size=args.min_component_size,
                               filtered_network_file=args.filtered_network)


def _test_topology_total_degree(args):
    test_topology(args.network, args.genesets, args.output,
        stat=st.geneset_total_degree_statistic,
        stat_name='topology_total_degree',
        setnames=args.setname, max_components=args.max_components,
        min_component_size=args.min_component_size,
        min_set_size=args.min_set_size,
        results_figure=args.results_figure,
        results_figure_type=args.results_figure_type,
        results_figure_significance=args.results_figure_significance,
        results_figure_x_label='Topology Total Degree',
        width=args.results_figure_width, height=args.results_figure_height,
        null_distributions=args.null_distributions,
        filtered_network_file=args.filtered_network,
        degree_correction_bins=args.degree_correction_bins,
        permutations=args.permutations, processes=args.cores)


def _test_topology_internal_degree(args):
    test_topology(args.network, args.genesets, args.output,
        setnames=args.setname, max_components=args.max_components,
        min_component_size=args.min_component_size,
        min_set_size=args.min_set_size,
        results_figure=args.results_figure,
        results_figure_type=args.results_figure_type,
        results_figure_significance=args.results_figure_significance,
        width=args.results_figure_width, height=args.results_figure_height,
        null_distributions=args.null_distributions,
        filtered_network_file=args.filtered_network,
        degree_correction_bins=args.degree_correction_bins,
        permutations=args.permutations, processes=args.cores)


def _test_topology_module(args):
    test_topology(args.network, args.genesets, args.output,
        stat=st.geneset_module_statistic,
        stat_name='topology_module',
        setnames=args.setname, max_components=args.max_components,
        min_component_size=args.min_component_size,
        min_set_size=args.min_set_size,
        results_figure=args.results_figure,
        results_figure_type=args.results_figure_type,
        results_figure_significance=args.results_figure_significance,
        results_figure_x_label='Largest Module Size',
        width=args.results_figure_width, height=args.results_figure_height,
        null_distributions=args.null_distributions,
        filtered_network_file=args.filtered_network,
        degree_correction_bins=args.degree_correction_bins,
        permutations=args.permutations, processes=args.cores)


def _test_topology_rwr(args):
    test_topology(args.network, args.genesets, args.output,
        rwr_matrix=args.rwr_matrix,
        stat=st.geneset_RW_statistic,
        stat_name='topology_rwr',
        setnames=args.setname, max_components=args.max_components,
        min_component_size=args.min_component_size,
        min_set_size=args.min_set_size,
        results_figure=args.results_figure,
        results_figure_type=args.results_figure_type,
        results_figure_significance=args.results_figure_significance,
        results_figure_x_label='Topology RWR statistic',
        width=args.results_figure_width, height=args.results_figure_height,
        null_distributions=args.null_distributions,
        filtered_network_file=args.filtered_network,
        rwr_matrix_in_memory=args.in_memory,
        degree_correction_bins=args.degree_correction_bins,
        permutations=args.permutations, processes=args.cores)


def _test_association(args):
    test_association(args.network, args.genesets_a, args.output,
        args.rwr_matrix, genesets_b=args.genesets_b,
        setnames_a=args.setname_a, setnames_b=args.setname_b,
        max_components=args.max_components,
        min_component_size=args.min_component_size,
        min_set_size=args.min_set_size,
        results_figure=args.results_figure,
        results_figure_type=args.results_figure_type,
        results_figure_significance=args.results_figure_significance,
        width=args.results_figure_width, height=args.results_figure_height,
        null_distributions=args.null_distributions,
        filtered_network_file=args.filtered_network,
        rwr_matrix_in_memory=args.in_memory,
        degree_correction_bins=args.degree_correction_bins,
        permutations=args.permutations, processes=args.cores)

def _plot_test_results(args):
    observed, results_type = parse_results_file(args.results)
    if results_type == 'one-v-many':
        if not args.type in {'box', 'strip', 'violin'}:
            raise RuntimeError('invalid figure type')
        if args.null_distributions:
            sep = '\t' if args.null_distributions.endswith('.tsv') else ','
            null_dists = pd.read_csv(args.null_distributions, sep=sep)
            plot_test_results(null_dists, observed.loc[:,['Gene set', 'Observed', 'Significance']], args.output,
                              figure_type=args.type,
                              significance=args.significance,
                              width=args.width, height=args.height,
                              x_label=ANALYSIS_LABEL[observed.loc[0,'Analysis']])
        else:
            raise RuntimeError('null distributions are required for one-v-many plot')
    else:
        figure_type = args.type.replace('box', 'obs')
        if not figure_type in {'obs', 'zscore', 'pvalue'}:
            raise RuntimeError('invalid figure type')
        observed['Z-score'] = (observed['Observed']-observed['Null mean'])/np.sqrt(observed['Null variance'])
        if results_type == 'all-v-all':
            plot_test_clustermap(observed.loc[:,['Gene set A', 'Gene set B', 'Observed', 'P-value', 'Z-score', 'Significance']],
                                 args.output, figure_type=figure_type,
                                 significance=args.significance,
                                 width=args.width, height=args.height)
        elif results_type == 'many-v-many':
            plot_test_heatmap(observed.loc[:,['Gene set A', 'Gene set B', 'Observed', 'P-value', 'Z-score', 'Significance']],
                                 args.output, figure_type=figure_type,
                                 significance=args.significance,
                                 width=args.width, height=args.height)
        


def add_network_arg(parser):
    parser.add_argument('network', metavar='<network.tsv>',
        help='Table defining network, 1st 2 cols should be gene pairs/edges')


def add_network_args(parser):
    add_network_arg(parser)
    parser.add_argument('--max-components', metavar='<int>', type=int,
        help='Use only n largest components')
    parser.add_argument('--min-component-size', metavar='<int>', type=int,
        default=2,
        help='Use only components with at least n genes (default: 2)')


def add_geneset_args(parser, full_arg=False):
    parser.add_argument('--genesets', metavar='<genesets.gmt>',
                        help='GMT file containing gene sets')
    parser.add_argument('--setname', metavar='<name>', nargs='+',
                        help='List of gene set names')
    parser.add_argument('--min-set-size', metavar='<int>', type=int, default=20,
                        help='Minimum gene set size (default: 20)')
    completeness_group = parser.add_mutually_exclusive_group()
    completeness_group.add_argument('--minimal', action='store_true',
        help='Filter out genes that are not part of a path between set members')
    completeness_group.add_argument('--strict', action='store_true',
        help='Filter out genes that are not part of a gene set')
    return completeness_group


def construct_network_parser(parser, func):
    parser.set_defaults(func=func)
    add_network_args(parser)
    return add_geneset_args(parser)


def add_output_component_args(parser):
    parser.add_argument('--max-output-components', metavar='<int>', type=int,
        help='Store only n largest components')
    parser.add_argument('--min-output-component-size', metavar='<int>',
        type=int, default=2,
        help='Store only components with at least n genes (default: 2)')


def add_results_fig_param_args(parser):
    parser.add_argument('--results-figure-significance', action='store_true',
                        help='Draw asterisks over significant results')
    parser.add_argument('--results-figure-width', metavar='<float>',
                        type=float, default=7.0,
                        help='Width of plot in inches')
    parser.add_argument('--results-figure-height', metavar='<float>',
                        type=float, default=7.0,
                        help='Height of plot in inches')


def add_null_dist_arg(parser):
    parser.add_argument('--null-distributions', metavar='<null.{csv,tsv}>',
                        help='Write null distributions to disk')

def add_filtered_network_arg(parser):
    parser.add_argument('--filtered-network', metavar='<filtered.tsv>',
        help='Write TSV file containing filtered network')


def add_permutation_args(parser):
    parser.add_argument('--degree-correction-bins', metavar='<int>', type=int,
        default=1, help='Apply degree correction using n bins (default: 1)')
    parser.add_argument('--permutations', metavar='<int>', type=int,
                        default=200,
                        help='Number of permutations (default: 200)')


def add_parallelization_args(parser):
    parser.add_argument('--cores', metavar='<int>', type=int, default=1,
                        help='Number of cores (default: 1)')


def construct_test_topology_parser(parser, func):
    parser.set_defaults(func=func)
    add_network_args(parser)
    parser.add_argument('genesets', metavar='<genesets.gmt>',
                        help='GMT file containing gene sets')
    parser.add_argument('output', metavar='<output.{csv,tsv}>',
                        help='Results table file')
    parser.add_argument('--setname', metavar='<name>', nargs='+',
                        help='List of gene set names')
    parser.add_argument('--min-set-size', metavar='<int>', type=int, default=20,
                        help='Minimum gene set size (default: 20)')
    parser.add_argument('--results-figure', metavar='<figure.{pdf,png,svg}>',
                        help='Path to results figure')
    parser.add_argument('--results-figure-type',
                        choices=('box', 'violin', 'strip'), default='box',
                        help='Type of results figure(default: box)')
    add_results_fig_param_args(parser)
    add_null_dist_arg(parser)
    add_filtered_network_arg(parser)
    add_permutation_args(parser)
    add_parallelization_args(parser)


def construct_test_association_parser(parser, func):
    parser.set_defaults(func=func)
    add_network_args(parser)
    parser.add_argument('genesets_a', metavar='<genesets-a.gmt>',
                        help='GMT file containing gene sets')
    parser.add_argument('output', metavar='<output.{csv,tsv}>',
                        help='Results table file')
    parser.add_argument('--genesets-b', metavar='<genesets-b.gmt>',
                        help='GMT file containing gene sets')
    parser.add_argument('--setname-a', metavar='<name>', nargs='+',
                        help='List of gene set names')
    parser.add_argument('--setname-b', metavar='<name>', nargs='+',
                        help='List of gene set names')
    parser.add_argument('--min-set-size', metavar='<int>', type=int, default=20,
                        help='Minimum gene set size (default: 20)')
    parser.add_argument('--results-figure', metavar='<figure.{pdf,png,svg}>',
                        help='Path to results figure')
    parser.add_argument('--results-figure-type',
                        choices=('box', 'violin', 'strip', 'obs', 'zscore',
                                 'pvalue'),
                        default='box',
                        help='Type of results figure (default: box/obs)')
    add_results_fig_param_args(parser)
    add_null_dist_arg(parser)
    add_filtered_network_arg(parser)
    add_permutation_args(parser)
    add_parallelization_args(parser)
    

def parse_arguments():
    parser = ArgumentParser(
        description='Statistical tests on gene networks and gene sets')
    parser.add_argument('--version', action='version',
        version='%(prog)s {version}'.format(version=__version__))
    parser.set_defaults(func=lambda _: parser.print_help(sys.stdout))
    subparsers = parser.add_subparsers()
    
    # Network parsers
    parser_network = subparsers.add_parser('network',
        help='Report properties of the network')
    network_parsers = parser_network.add_subparsers()
    parser_network_summary = network_parsers.add_parser('summary',
        help='Summarize the network')
    construct_network_parser(parser_network_summary, _summary)
    parser_network_summary.add_argument('output', metavar='<output.txt>',
                              help='Text file containing summary')
    parser_network_summary.add_argument('--network-name', metavar='<name>',
        default='network', help='A name for the network')
    parser_network_summary.add_argument('--plot-degree',
        metavar='<output.{pdf,png}>', help='Plot the degree distribution')
    parser_network_summary.add_argument('--plot-component',
        metavar='<output.{pdf,png}>',
        help='Plot the distribution of component sizes')
    parser_network_summary.add_argument('--filtered-network',
        metavar='<filtered.tsv>',
        help='Write TSV file containing filtered network')
    parser_network_filter = network_parsers.add_parser('filter',
        help='Filter small components out of the network')
    construct_network_parser(parser_network_filter, _filter_network)
    add_output_component_args(parser_network_filter)
    parser_network_filter.add_argument('output', metavar='<output.tsv>',
        help='TSV file containing filtered network')
    parser_network_cytoscape = network_parsers.add_parser('cytoscape',
        help='Generate a GraphML file that can be viewed in cytoscape')
    completeness_group_cytoscape = construct_network_parser(parser_network_cytoscape, _cytoscape)
    add_output_component_args(parser_network_cytoscape)
    parser_network_cytoscape.add_argument('output_graphml',
        metavar='<output.graphml>',
        help='GraphML file containing network or subnetwork')
    parser_network_cytoscape.add_argument('--components-out',
        metavar='<components.gmt>',
        help='Write GMT file containing components of the output graph')
    completeness_group_cytoscape.add_argument('--full', action='store_true',
        help='Store the full network')
    parser_network_cytoscape.add_argument('--filtered-network',
        metavar='<filtered.tsv>',
        help='Write TSV file containing filtered network')
    
    # Build parser
    parser_build = subparsers.add_parser('build',
        help='Build a RWR diffusion matrix')
    parser_build.set_defaults(func=_build_rwr_diffusion_matrix)
    add_network_args(parser_build)
    parser_build.add_argument('output', metavar='<rwr-matrix.hdf5>',
                         help='Output file for RWR diffusion matrix')
    parser_build.add_argument('--beta', metavar='<float>', type=float,
                         default=0.85, help='Restart probability (default: 0.85)')
    parser_build.add_argument('--filtered-network', metavar='<filtered.tsv>',
                              help='Write TSV file containing filtered network')

    # Test parsers
    parser_test = subparsers.add_parser('test', help='Statistical tests')
    test_parsers = parser_test.add_subparsers()
    parser_ttd = test_parsers.add_parser('ttd',
        help='Topology Total Degree statistic')
    construct_test_topology_parser(parser_ttd, _test_topology_total_degree)
    parser_tid = test_parsers.add_parser('tid',
        help='Topology Internal Degree statistic')
    construct_test_topology_parser(parser_tid, _test_topology_internal_degree)
    parser_tm = test_parsers.add_parser('tm',
        help='Topology Module statistic (largest module size)')
    construct_test_topology_parser(parser_tm, _test_topology_module)
    parser_trwr = test_parsers.add_parser('trwr',
        help='Topology Random Walk with Restart diffusion statistic')
    construct_test_topology_parser(parser_trwr, _test_topology_rwr)
    parser_trwr.add_argument('rwr_matrix', metavar='<rwr-matrix.hdf5>',
        help='RWR diffusion matrix')
    parser_trwr.add_argument('--in-memory', action='store_true',
        help='Load the entire RWR diffusion matrix in memory')
    parser_arwr = test_parsers.add_parser('arwr',
        help='Association Random Walk with Restart diffusion statistic')
    construct_test_association_parser(parser_arwr, _test_association)
    parser_arwr.add_argument('rwr_matrix', metavar='<rwr-matrix.hdf5>',
        help='RWR diffusion matrix')
    parser_arwr.add_argument('--in-memory', action='store_true',
        help='Load the entire RWR diffusion matrix in memory')

    # Plot parsers
    parser_plot = subparsers.add_parser('plot',
        help='Plot a figure from a results table')
    parser_plot.set_defaults(func=_plot_test_results)
    parser_plot.add_argument('results', metavar='<results.{csv,tsv}>',
                        help='Results table from pygna2 test')
    parser_plot.add_argument('output', metavar='<output.{pdf,png,svg}>',
                        help='Output file for plot')
    parser_plot.add_argument('--null-distributions',
                        metavar='<null_distributions.{csv,tsv}>',
                        help='Null distributions from pygna2 test, required for topology or one-v-many')
    parser_plot.add_argument('--type', choices=('box', 'violin', 'strip', 'obs',
                                           'zscore', 'pvalue'),
                        default='box',
                        help='Type of results figure (default: box/obs)')
    parser_plot.add_argument('--significance', action='store_true',
                        help='Draw asterisks over significant results')
    parser_plot.add_argument('--width', metavar='<float>', type=float, default=7.0,
                        help='Width of plot in inches')
    parser_plot.add_argument('--height', metavar='<float>', type=float, default=7.0,
                        help='Height of plot in inches')
    return parser.parse_args()

def main():
    args = parse_arguments()
    args.func(args)
